# pylint: disable=no-self-use,invalid-name,protected-access

from typing import Dict, List, Tuple, Any
from collections import OrderedDict

import torch

from allennlp.common.checks import ConfigurationError
from allennlp.training.optimizers import Optimizer
from allennlp.common.testing import AllenNlpTestCase
from allennlp.training.learning_rate_schedulers import LearningRateScheduler
from allennlp.common.params import Params


class LearningRateSchedulersTest(AllenNlpTestCase):

    def test_reduce_on_plateau_error_throw_when_no_metrics_exist(self):
        model = torch.nn.Sequential(torch.nn.Linear(10, 10))
        with self.assertRaises(ConfigurationError) as context:
            LearningRateScheduler.from_params(Optimizer.from_params(model.named_parameters(),
                                                                    Params({"type": "adam"})),
                                              Params({"type": "reduce_on_plateau"})).step(None, None)

        self.assertTrue(
                'The reduce_on_plateau learning rate scheduler requires a validation metric'
                in str(context.exception))

    def test_reduce_on_plateau_works_when_metrics_exist(self):
        model = torch.nn.Sequential(torch.nn.Linear(10, 10))
        LearningRateScheduler.from_params(Optimizer.from_params(model.named_parameters(),
                                                                Params({"type": "adam"})),
                                          Params({"type": "reduce_on_plateau"})).step(10, None)

    def test_no_metric_wrapper_can_support_none_for_metrics(self):
        model = torch.nn.Sequential(torch.nn.Linear(10, 10))
        lrs = LearningRateScheduler.from_params(Optimizer.from_params(model.named_parameters(),
                                                                      Params({"type": "adam"})),
                                                Params({"type": "step", "step_size": 1}))
        lrs.step(None, None)

    def test_noam_learning_rate_schedule_does_not_crash(self):
        model = torch.nn.Sequential(torch.nn.Linear(10, 10))
        lrs = LearningRateScheduler.from_params(Optimizer.from_params(model.named_parameters(),
                                                                      Params({"type": "adam"})),
                                                Params({"type": "noam", "model_size": 10, "warmup_steps": 2000}))
        lrs.step(None)
        lrs.step_batch(None)


class CosineWithRestartsTest(AllenNlpTestCase):

    def setUp(self):
        super(CosineWithRestartsTest, self).setUp()
        self.model = torch.nn.Sequential(torch.nn.Linear(10, 10))

        # We use these cases to verify that the scheduler works as expected.
        # Each case consists of 5 parameters:
        # - epochs: the total # of epochs to run for.
        # - params: parameters passed to initialize the scheduler.
        # - learning rate checks: a list of tuples, each of which specifies an epoch
        #   number and the expected value of the learning rate at that epoch.
        # - checkpoints: a list of epoch numbers at which to save the scheduler
        #   state, and then restore from the saved state and resume.
        self.cosine_schedule_cases = [
                (
                        30,
                        {"t_initial": 30, "t_mul": 1.0},
                        [(0, 1.0),
                         (15, 0.5000000000000001),
                         (29, 0.0027390523158632996)],
                        [10, 14]
                ),
                (
                        10,
                        {"t_initial": 1, "t_mul": 2.0},
                        [(0, 1.0),
                         (1, 1.0),
                         (2, 0.5),
                         (3, 1.0)],
                        [1, 3]
                ),
                (
                        30,
                        {"t_initial": 1, "t_mul": 1.0},
                        [(0, 1.0),
                         (15, 1.0),
                         (29, 1.0)],
                        []
                ),
                (
                        60,
                        {"t_initial": 30, "t_mul": 1.0},
                        [(0, 1.0),
                         (15, 0.5000000000000001),
                         (29, 0.0027390523158632996),
                         (30, 1.0),
                         (45, 0.5000000000000001),
                         (59, 0.0027390523158632996)],
                        [30, 35]
                ),
                (
                        60,
                        {"t_initial": 30, "t_mul": 1.0, "eta_mul": 0.5},
                        [(0, 1.0),
                         (15, 0.5000000000000001),
                         (29, 0.0027390523158632996),
                         (30, .5)],
                        []
                ),
                (
                        100,
                        {"t_initial": 30, "t_mul": 1.5},
                        [(0, 1.0),
                         (29, 0.0027390523158632996),
                         (30, 1.0),
                         (74, 0.0012179748700879012)],
                        []
                ),
                (
                        210,
                        {"t_initial": 30, "t_mul": 2},
                        [(0, 1.0),
                         (29, 0.0027390523158632996),
                         (30, 1.0),
                         (89, 0.0006852326227130834),
                         (90, 1.0),
                         (209, 0.00017133751222137006)],
                        []
                ),
                (
                        210,
                        {"t_initial": 30, "t_mul": 2, "eta_mul": 0.5},
                        [(0, 1.0),
                         (30, .5),
                         (90, .25)],
                        [29, 90]
                ),
                (
                        150,
                        {"t_initial": 30, "t_mul": 1},
                        [(0, 1.0),
                         (29, 0.0027390523158632996),
                         (30, 1.0),
                         (59, 0.0027390523158632996),
                         (60, 1.0),
                         (89, 0.0027390523158632996),
                         (90, 1.0)],
                        []
                ),
                (
                        10,
                        {"t_initial": 1, "t_mul": 1, "eta_mul": 0.5},
                        [(0, 1.0), (1, 0.5), (2, 0.25)],
                        []
                ),
        ]

    def _get_optimizer(self, lr: float = 1.0):
        return Optimizer.from_params(self.model.named_parameters(), Params({"type": "sgd", "lr": lr}))

    def test_from_params(self):
        """Make sure ``from_params`` initializes an instance properly."""
        optim = self._get_optimizer()
        sched = LearningRateScheduler.from_params(optim, Params({"type": "cosine", "t_initial": 5})).lr_scheduler

        assert sched.t_initial == 5
        assert sched._initialized is True

        # Learning should be unchanged after initializing scheduler.
        assert optim.param_groups[0]["lr"] == 1.0

        with self.assertRaises(TypeError):
            # t_initial is required.
            LearningRateScheduler.from_params(optim, Params({"type": "cosine"}))

    def test_schedules(self):
        """Make sure the math is correct."""
        for epochs, params, lr_checks, _ in self.cosine_schedule_cases:
            optimizer = self._get_optimizer()
            params["type"] = "cosine"
            scheduler = LearningRateScheduler.from_params(optimizer, Params(params)).lr_scheduler
            lrs = [optimizer.param_groups[0]["lr"]]
            for epoch in range(epochs):
                scheduler.step(epoch)
                lrs.append(optimizer.param_groups[0]["lr"])

            for it, lr in lr_checks:
                assert lrs[it] == lr

    def test_schedules_with_save_and_resume(self):
        """Make sure scheduler will resume with the right state."""

        def init_and_restore_scheduler(optimizer: torch.optim.Optimizer,
                                       params: Dict[str, Any],
                                       state_dict: Dict[str, Any] = None):
            """
            Initialize a new scheduler and optionally restore its state from
            a checkpoint.
            """
            params["type"] = "cosine"
            scheduler = LearningRateScheduler.from_params(optimizer, Params(params))
            if state_dict is not None:
                scheduler.lr_scheduler.load_state_dict(state_dict)
            return scheduler

        for epochs, params, lr_checks, checkpoints in self.cosine_schedule_cases:
            optimizer = self._get_optimizer()
            scheduler = init_and_restore_scheduler(optimizer, params)
            state = scheduler.lr_scheduler.state_dict()

            lrs = [optimizer.param_groups[0]["lr"]]
            for epoch in range(epochs):
                if epoch in checkpoints:
                    # Restore scheduler from state dict.
                    scheduler = init_and_restore_scheduler(optimizer, params, state_dict=state)

                # Take step and record learning rate.
                scheduler.step(1, epoch)
                lrs.append(optimizer.param_groups[0]["lr"])

                # Save state again.
                state = scheduler.lr_scheduler.state_dict()

            for it, lr in lr_checks:
                assert lrs[it] == lr


def is_hat_shaped(learning_rates: List[float]):
    """
    Check if the list of learning rates is "hat" shaped, i.e.,
    increases then decreases
    """
    # sufficient conditions:
    #   has both an increasing and decreasing segment
    #   decrease segment occurs after increasing segment
    #   once start decreasing, can't increase again
    has_increasing_segment = False
    has_decreasing_segment = False
    for k in range(1, len(learning_rates)):
        delta = learning_rates[k] - learning_rates[k-1]
        if delta > 1e-8:
            has_increasing_segment = True
            if has_decreasing_segment:
                # can't increase again after hitting the max
                return False
        elif delta < -1e-8:
            if not has_increasing_segment:
                # can't decrease without have an increasing segment
                return False
            has_decreasing_segment = True
        else:
            # no change
            pass

    return has_increasing_segment and has_decreasing_segment


class SlantedTriangularTest(AllenNlpTestCase):

    def setUp(self):
        super(SlantedTriangularTest, self).setUp()
        self.model = torch.nn.Sequential(OrderedDict([
                ('lin1', torch.nn.Linear(10, 10)),
                ('lin2', torch.nn.Linear(10, 10))
        ]))

    def _get_optimizer(self, lr: float = 1.0):
        optimizer_params = Params({"type": "sgd", "lr": lr})
        optimizer_params["parameter_groups"] = [[[f"^{m}"], {}] for m in self.model._modules]
        return Optimizer.from_params(self.model.named_parameters(), optimizer_params)

    def _run_scheduler_get_lrs(self, params, num_steps_per_epoch):
        optimizer = self._get_optimizer()
        params["type"] = "slanted_triangular"
        scheduler = LearningRateScheduler.from_params(optimizer, Params(params))
        lrs = []

        batch_num_total = 0
        for epoch in range(params["num_epochs"]):
            for _ in range(num_steps_per_epoch):
                batch_num_total += 1
                # allennlp trainer calls step_batch after updating parameters
                # so collect lr at time of parameter update
                lrs.append([param_group["lr"] * float(param_group['params'][0].requires_grad)
                            for param_group in optimizer.param_groups[:2]])
                scheduler.step_batch(batch_num_total)
                if params.get("gradual_unfreezing") and epoch == 0:
                    assert scheduler.lr_scheduler.freezing_current
            # step() takes two arguments: validation metric and epoch
            scheduler.step(None, epoch)

        return lrs

    def test_is_hat_shaped(self):
        assert not is_hat_shaped([0.0] * 10)
        assert not is_hat_shaped([float(k) for k in range(10)])
        assert not is_hat_shaped([float(10 - k) for k in range(10)])
        assert is_hat_shaped([float(k) for k in range(10)] +
                             [float(10 - k) for k in range(10)])
        assert not is_hat_shaped([float(k) for k in range(10)] +
                                 [float(10 - k) for k in range(10)] +
                                 [float(k) for k in range(10)])

    def test_from_params(self):
        optim = self._get_optimizer()
        sched = LearningRateScheduler.from_params(
                optim, Params({"type": "slanted_triangular",
                               "num_epochs": 5,
                               "num_steps_per_epoch": 10,
                               "gradual_unfreezing": True,
                               "discriminative_fine_tuning": True,
                               "decay_factor": 0.5})).lr_scheduler

        assert sched.num_epochs == 5
        assert sched.num_steps_per_epoch == 10
        assert sched.gradual_unfreezing is True
        assert sched.freezing_current is True

        assert len(optim.param_groups) == 3
        # The default parameter group in the Optimizer is empty
        assert not optim.param_groups[-1]["params"]
        assert optim.param_groups[-2]["lr"] == 1.0 / sched.ratio
        assert optim.param_groups[-3]["lr"] == 0.5 / sched.ratio

        with self.assertRaises(TypeError):
            # num_epochs and num_steps_per_epoch are required
            LearningRateScheduler.from_params(optim, Params({"type": "slanted_triangular", "num_epochs": 5}))
            LearningRateScheduler.from_params(
                    optim,
                    Params({"type": "slanted_triangular", "num_steps_epochs": 10}))

    def test_schedules(self):
        slanted_triangular_cases: List[Tuple[Dict[str, Any], List[Tuple[int, float]]]] = [
                (
                        {"num_epochs": 5,
                         "num_steps_per_epoch": 10,
                         "gradual_unfreezing": True},  # parameters
                        [(0, 1, 0.03125),  # iteration, layer, learning rate
                         (0, 0, 0.0),
                         (1, 1, 1.0),
                         (1, 0, 0.0),
                         (9, 1, 0.138888),
                         (9, 0, 0.0),      # end of the first epoch
                         (10, 1, 0.03125),
                         (10, 0, 0.03125),
                         (14, 1, 1.0),
                         (14, 0, 1.0),
                         (49, 1, 0.05815972),
                         (49, 0, 0.05815972)]
                ),
                (
                        {"num_epochs": 5,
                         "num_steps_per_epoch": 10,
                         "discriminative_fine_tuning": True,
                         "decay_factor": 0.5},  # parameters
                        [(0, 1, 0.03125),  # iteration, layer, learning rate
                         (0, 0, 0.015625),
                         (5, 1, 1.0),
                         (5, 0, 0.5),
                         (49, 1, 0.052777),
                         (49, 0, 0.026388)]
                ),
                (
                        {"num_epochs": 5,
                         "num_steps_per_epoch": 10,
                         "gradual_unfreezing": True,
                         "discriminative_fine_tuning": True,
                         "decay_factor": 0.5},  # parameters
                        [(0, 1, 0.03125),  # iteration, layer, learning rate
                         (0, 0, 0.0),
                         (1, 1, 1.0),
                         (1, 0, 0.0),
                         (9, 1, 0.138888),
                         (9, 0, 0.0),      # end of the first epoch
                         (10, 1, 0.03125),
                         (10, 0, 0.015625),
                         (14, 1, 1.0),
                         (14, 0, 0.5),
                         (49, 1, 0.0581597222),
                         (49, 0, 0.0290798611)]
                )
        ]
        for params, lr_checks in slanted_triangular_cases:
            lrs = self._run_scheduler_get_lrs(params, params["num_steps_per_epoch"])

            for it, layer, lr in lr_checks:
                lr_check = round(lr, 5)
                lr = round(lrs[it][layer], 5)
                assert lr == lr_check, f'Learning rate {lr} at iteration {it} at layer {layer} != {lr_check}.'

    def test_schedules_num_steps_per_epoch(self):
        # ensure the learning rate schedule still maintains hat shape
        # if number of actual batches differs from parameter provided
        # in constructor
        for gradual_unfreezing in [True, False]:
            for discriminative_fine_tuning in [True, False]:
                for num_actual_steps_per_epoch in [7, 11]:
                    params = {"num_epochs": 5,
                              "num_steps_per_epoch": 10,
                              "gradual_unfreezing": gradual_unfreezing,
                              "discriminative_fine_tuning": discriminative_fine_tuning}
                    lrs = self._run_scheduler_get_lrs(params, num_actual_steps_per_epoch)
                    first_layer_lrs = [rates[0] for rates in lrs]
                    second_layer_lrs = [rates[1] for rates in lrs]

                    if gradual_unfreezing:
                        assert max(first_layer_lrs[:num_actual_steps_per_epoch]) < 1e-8
                        assert min(first_layer_lrs[:num_actual_steps_per_epoch]) > -1e-8
                        assert is_hat_shaped(first_layer_lrs[num_actual_steps_per_epoch:])
                        assert is_hat_shaped(second_layer_lrs[:num_actual_steps_per_epoch])
                        assert is_hat_shaped(second_layer_lrs[num_actual_steps_per_epoch:])
                    else:
                        assert is_hat_shaped(first_layer_lrs)
                        assert is_hat_shaped(second_layer_lrs)



# pylint: disable=no-self-use,invalid-name,protected-access
import json

from overrides import overrides

from allennlp.common.testing import AllenNlpTestCase
from allennlp.data.tokenizers import Token
from allennlp.semparse import ParsingError, World
from allennlp.semparse.contexts import TableQuestionKnowledgeGraph
from allennlp.semparse.worlds import NlvrWorld, WikiTablesWorld


class FakeWorldWithoutRecursion(World):
    # pylint: disable=abstract-method
    @overrides
    def all_possible_actions(self):
        # The logical forms this grammar allows are
        # (unary_function argument)
        # (binary_function argument argument)
        actions = ['@start@ -> t',
                   't -> [<e,t>, e]',
                   '<e,t> -> unary_function',
                   '<e,t> -> [<e,<e,t>>, e]',
                   '<e,<e,t>> -> binary_function',
                   'e -> argument']
        return actions


class FakeWorldWithRecursion(FakeWorldWithoutRecursion):
    # pylint: disable=abstract-method
    @overrides
    def all_possible_actions(self):
        # In addition to the forms allowed by ``FakeWorldWithoutRecursion``, this world allows
        # (unary_function (identity .... (argument)))
        # (binary_function (identity .... (argument)) (identity .... (argument)))
        actions = super(FakeWorldWithRecursion, self).all_possible_actions()
        actions.extend(['e -> [<e,e>, e]',
                        '<e,e> -> identity'])
        return actions


class TestWorld(AllenNlpTestCase):
    def setUp(self):
        super().setUp()
        self.world_without_recursion = FakeWorldWithoutRecursion()
        self.world_with_recursion = FakeWorldWithRecursion()

        test_filename = self.FIXTURES_ROOT / "data" / "nlvr" / "sample_ungrouped_data.jsonl"
        data = [json.loads(line)["structured_rep"] for line in open(test_filename).readlines()]
        self.nlvr_world = NlvrWorld(data[0])

        question_tokens = [Token(x) for x in ['what', 'was', 'the', 'last', 'year', '2004', '?']]
        table_file = self.FIXTURES_ROOT / 'data' / 'wikitables' / 'sample_table.tsv'
        table_kg = TableQuestionKnowledgeGraph.read_from_file(table_file, question_tokens)
        self.wikitables_world = WikiTablesWorld(table_kg)

    def test_get_paths_to_root_without_recursion(self):
        argument_paths = self.world_without_recursion.get_paths_to_root('e -> argument')
        assert argument_paths == [['e -> argument', 't -> [<e,t>, e]', '@start@ -> t'],
                                  ['e -> argument', '<e,t> -> [<e,<e,t>>, e]', 't -> [<e,t>, e]',
                                   '@start@ -> t']]
        unary_function_paths = self.world_without_recursion.get_paths_to_root('<e,t> -> unary_function')
        assert unary_function_paths == [['<e,t> -> unary_function', 't -> [<e,t>, e]',
                                         '@start@ -> t']]
        binary_function_paths = \
                self.world_without_recursion.get_paths_to_root('<e,<e,t>> -> binary_function')
        assert binary_function_paths == [['<e,<e,t>> -> binary_function',
                                          '<e,t> -> [<e,<e,t>>, e]', 't -> [<e,t>, e]',
                                          '@start@ -> t']]

    def test_get_paths_to_root_with_recursion(self):
        argument_paths = self.world_with_recursion.get_paths_to_root('e -> argument')
        # Argument now has 4 paths, and the two new paths are with the identity function occurring
        # (only once) within unary and binary functions.
        assert argument_paths == [['e -> argument', 't -> [<e,t>, e]', '@start@ -> t'],
                                  ['e -> argument', '<e,t> -> [<e,<e,t>>, e]', 't -> [<e,t>, e]',
                                   '@start@ -> t'],
                                  ['e -> argument', 'e -> [<e,e>, e]', 't -> [<e,t>, e]',
                                   '@start@ -> t'],
                                  ['e -> argument', 'e -> [<e,e>, e]', '<e,t> -> [<e,<e,t>>, e]',
                                   't -> [<e,t>, e]', '@start@ -> t']]
        identity_paths = self.world_with_recursion.get_paths_to_root('<e,e> -> identity')
        # Two identity paths, one through each of unary and binary function productions.
        assert identity_paths == [['<e,e> -> identity', 'e -> [<e,e>, e]', 't -> [<e,t>, e]',
                                   '@start@ -> t'],
                                  ['<e,e> -> identity', 'e -> [<e,e>, e]',
                                   '<e,t> -> [<e,<e,t>>, e]', 't -> [<e,t>, e]', '@start@ -> t']]

    # The tests for get_action_sequence and get_logical_form need a concrete world to be useful;
    # we'll mostly use the NLVR world to test them, as it's a simpler world than the WikiTables
    # world.

    def test_get_action_sequence_removes_currying(self):
        world = self.wikitables_world
        logical_form = ("(argmax (number 1) (number 1) (fb:row.row.division fb:cell.2) "
                        "(reverse (lambda x ((reverse fb:row.row.index) (var x))))")
        parsed_logical_form = world.parse_logical_form(logical_form)
        action_sequence = world.get_action_sequence(parsed_logical_form)
        assert 'r -> [<n,<n,<#1,<<#2,#1>,#1>>>>, n, n, r, <n,r>]' in action_sequence

    def test_get_action_sequence_removes_and_retains_var_correctly(self):
        world = self.wikitables_world
        logical_form = ("((reverse fb:row.row.league) (argmin (number 1) (number 1) "
                        "(fb:type.object.type fb:type.row) "
                        "(reverse (lambda x ((reverse fb:row.row.index) (var x))))))")
        parsed_logical_form_without_var = world.parse_logical_form(logical_form)
        action_sequence_without_var = world.get_action_sequence(parsed_logical_form_without_var)
        assert '<#1,#1> -> var' not in action_sequence_without_var

        parsed_logical_form_with_var = world.parse_logical_form(logical_form,
                                                                remove_var_function=False)
        action_sequence_with_var = world.get_action_sequence(parsed_logical_form_with_var)
        assert '<#1,#1> -> var' in action_sequence_with_var

    def test_get_logical_form_handles_reverse(self):
        world = self.wikitables_world
        logical_form = "((reverse fb:row.row.year) (fb:row.row.league fb:cell.usl_a_league))"
        parsed_logical_form = world.parse_logical_form(logical_form)
        action_sequence = world.get_action_sequence(parsed_logical_form)
        reconstructed_logical_form = world.get_logical_form(action_sequence)
        parsed_reconstructed_logical_form = world.parse_logical_form(reconstructed_logical_form)
        assert parsed_logical_form == parsed_reconstructed_logical_form

        logical_form = ("((reverse fb:cell.cell.date) ((reverse fb:row.row.year) (argmax (number 1) "
                        "(number 1) (fb:row.row.league fb:cell.usl_a_league) (reverse (lambda x "
                        "((reverse fb:row.row.index) (var x)))))))")
        parsed_logical_form = world.parse_logical_form(logical_form)
        action_sequence = world.get_action_sequence(parsed_logical_form)
        reconstructed_logical_form = world.get_logical_form(action_sequence)
        parsed_reconstructed_logical_form = world.parse_logical_form(reconstructed_logical_form)
        assert parsed_logical_form == parsed_reconstructed_logical_form

    def test_get_logical_form_handles_greater_than(self):
        world = self.wikitables_world
        action_sequence = ['@start@ -> c', 'c -> [<r,c>, r]', '<r,c> -> [<<#1,#2>,<#2,#1>>, <c,r>]',
                           '<<#1,#2>,<#2,#1>> -> reverse', '<c,r> -> fb:row.row.league',
                           'r -> [<c,r>, c]', '<c,r> -> fb:row.row.year', 'c -> [<n,c>, n]',
                           '<n,c> -> fb:cell.cell.number', 'n -> [<nd,nd>, n]', '<nd,nd> -> >',
                           'n -> [<n,n>, n]', '<n,n> -> number', 'n -> 2004']
        logical_form = world.get_logical_form(action_sequence)
        expected_logical_form = ('((reverse fb:row.row.league) (fb:row.row.year '
                                 '(fb:cell.cell.number (> (number 2004)))))')
        assert logical_form == expected_logical_form

    def test_get_logical_form_handles_length_one_terminal_functions(self):
        world = self.wikitables_world
        logical_form = ("(- ((reverse fb:cell.cell.number) ((reverse fb:row.row.league) "
                        "(fb:row.row.year fb:cell.usl_a_league))) (number 1))")
        parsed_logical_form = world.parse_logical_form(logical_form)
        action_sequence = world.get_action_sequence(parsed_logical_form)
        reconstructed_logical_form = world.get_logical_form(action_sequence)
        parsed_reconstructed_logical_form = world.parse_logical_form(reconstructed_logical_form)
        assert parsed_logical_form == parsed_reconstructed_logical_form

    def test_get_logical_form_with_real_logical_forms(self):
        nlvr_world = self.nlvr_world
        logical_form = ("(box_count_greater_equals (member_color_count_equals all_boxes 1) 1)")
        parsed_logical_form = nlvr_world.parse_logical_form(logical_form)
        action_sequence = nlvr_world.get_action_sequence(parsed_logical_form)
        reconstructed_logical_form = nlvr_world.get_logical_form(action_sequence)
        parsed_reconstructed_logical_form = nlvr_world.parse_logical_form(reconstructed_logical_form)
        # It makes more sense to compare parsed logical forms instead of actual logical forms.
        assert parsed_logical_form == parsed_reconstructed_logical_form
        assert nlvr_world.execute(logical_form) == nlvr_world.execute(reconstructed_logical_form)
        logical_form = "(object_color_all_equals (circle (touch_wall (all_objects))) color_black)"
        parsed_logical_form = nlvr_world.parse_logical_form(logical_form)
        action_sequence = nlvr_world.get_action_sequence(parsed_logical_form)
        reconstructed_logical_form = nlvr_world.get_logical_form(action_sequence)
        parsed_reconstructed_logical_form = nlvr_world.parse_logical_form(reconstructed_logical_form)
        assert parsed_logical_form == parsed_reconstructed_logical_form
        assert nlvr_world.execute(logical_form) == nlvr_world.execute(reconstructed_logical_form)

    def test_get_logical_form_fails_with_incomplete_action_sequence(self):
        nlvr_world = self.nlvr_world
        action_sequence = ['@start@ -> t', 't -> [<b,t>, b]', '<b,t> -> box_exists']
        with self.assertRaisesRegex(ParsingError, 'Incomplete action sequence'):
            nlvr_world.get_logical_form(action_sequence)

    def test_get_logical_form_fails_with_extra_actions(self):
        nlvr_world = self.nlvr_world
        action_sequence = ['@start@ -> <b,t>', '<b,t> -> box_exists', 't -> [<b,t>, b]']
        with self.assertRaisesRegex(ParsingError, 'Extra actions'):
            nlvr_world.get_logical_form(action_sequence)

    def test_get_logical_form_fails_with_action_sequence_in_wrong_order(self):
        nlvr_world = self.nlvr_world
        action_sequence = ['@start@ -> t', 't -> [<b,t>, b]', '<b,t> -> box_exists',
                           'b -> [<c,b>, c]', '<c,b> -> [<b,<c,b>>, b]',
                           'b -> all_boxes', '<b,<c,b>> -> member_color_none_equals',
                           'c -> color_blue']
        with self.assertRaisesRegex(ParsingError, 'does not match'):
            nlvr_world.get_logical_form(action_sequence)

    def test_get_logical_form_adds_var_correctly(self):
        world = self.wikitables_world
        action_sequence = ['@start@ -> e', 'e -> [<r,e>, r]', '<r,e> -> [<<#1,#2>,<#2,#1>>, <e,r>]',
                           '<<#1,#2>,<#2,#1>> -> reverse', '<e,r> -> fb:row.row.league',
                           'r -> [<d,<d,<#1,<<d,#1>,#1>>>>, d, d, r, <d,r>]',
                           '<d,<d,<#1,<<d,#1>,#1>>>> -> argmin', 'd -> [<e,d>, e]', '<e,d> -> number',
                           'e -> 1', 'd -> [<e,d>, e]', '<e,d> -> number', 'e -> 1',
                           'r -> [<#1,#1>, r]', '<#1,#1> -> fb:type.object.type', 'r -> fb:type.row',
                           '<d,r> -> [<<#1,#2>,<#2,#1>>, <r,d>]', '<<#1,#2>,<#2,#1>> -> reverse',
                           "<r,d> -> ['lambda x', d]", 'd -> [<r,d>, r]',
                           '<r,d> -> [<<#1,#2>,<#2,#1>>, <d,r>]', '<<#1,#2>,<#2,#1>> -> reverse',
                           '<d,r> -> fb:row.row.index', 'r -> x']
        logical_form = world.get_logical_form(action_sequence)
        assert '(var x)' in logical_form
        expected_logical_form = ("((reverse fb:row.row.league) (argmin (number 1) (number 1) "
                                 "(fb:type.object.type fb:type.row) "
                                 "(reverse (lambda x ((reverse fb:row.row.index) (var x))))))")
        parsed_logical_form = world.parse_logical_form(logical_form)
        parsed_expected_logical_form = world.parse_logical_form(expected_logical_form)
        assert parsed_logical_form == parsed_expected_logical_form

    def test_get_logical_form_fails_with_unnecessary_add_var(self):
        world = self.wikitables_world
        action_sequence = ['@start@ -> e', 'e -> [<r,e>, r]', '<r,e> -> [<<#1,#2>,<#2,#1>>, <e,r>]',
                           '<<#1,#2>,<#2,#1>> -> reverse', '<e,r> -> fb:row.row.league',
                           'r -> [<d,<d,<#1,<<d,#1>,#1>>>>, d, d, r, <d,r>]',
                           '<d,<d,<#1,<<d,#1>,#1>>>> -> argmin', 'd -> [<e,d>, e]', '<e,d> -> number',
                           'e -> 1', 'd -> [<e,d>, e]', '<e,d> -> number', 'e -> 1',
                           'r -> [<#1,#1>, r]', '<#1,#1> -> fb:type.object.type', 'r -> fb:type.row',
                           '<d,r> -> [<<#1,#2>,<#2,#1>>, <r,d>]', '<<#1,#2>,<#2,#1>> -> reverse',
                           "<r,d> -> ['lambda x', d]", 'd -> [<r,d>, r]',
                           '<r,d> -> [<<#1,#2>,<#2,#1>>, <d,r>]', '<<#1,#2>,<#2,#1>> -> reverse',
                           '<d,r> -> fb:row.row.index', 'r -> [<#1,#1>, r]', '<#1,#1> -> var', 'r -> x']
        with self.assertRaisesRegex(ParsingError, 'already had var'):
            world.get_logical_form(action_sequence)

    def test_get_logical_form_with_multiple_negate_filters(self):
        world = self.nlvr_world
        # This is an actual sequence of actions produced by an untrained NlvrSemanticParser
        action_sequence = ['@start@ -> t', 't -> [<o,<c,t>>, o, c]',
                           '<o,<c,t>> -> object_color_all_equals', 'o -> [<o,o>, o]',
                           '<o,o> -> [<<o,o>,<o,o>>, <o,o>]', '<<o,o>,<o,o>> -> negate_filter',
                           '<o,o> -> [<<o,o>,<o,o>>, <o,o>]', '<<o,o>,<o,o>> -> negate_filter',
                           '<o,o> -> blue', 'o -> [<o,o>, o]', '<o,o> -> blue', 'o -> [<o,o>, o]',
                           '<o,o> -> blue', 'o -> all_objects', 'c -> color_blue']
        logical_form = world.get_logical_form(action_sequence)
        # "The color of all blue blue objects that are not not blue is blue".
        expected_logical_form = ("(object_color_all_equals ((negate_filter (negate_filter blue)) "
                                 "(blue (blue all_objects))) color_blue)")
        parsed_logical_form = world.parse_logical_form(logical_form)
        parsed_expected_logical_form = world.parse_logical_form(expected_logical_form)
        assert parsed_logical_form == parsed_expected_logical_form



# pylint: disable=no-self-use,invalid-name
import json

from allennlp.common.testing import AllenNlpTestCase
from allennlp.semparse.worlds.world import ExecutionError
from allennlp.semparse.executors import NlvrExecutor
from allennlp.semparse.worlds.nlvr_box import Box


class TestNlvrExecutor(AllenNlpTestCase):
    def setUp(self):
        super().setUp()
        test_filename = self.FIXTURES_ROOT / "data" / "nlvr" / "sample_ungrouped_data.jsonl"
        data = [json.loads(line)["structured_rep"] for line in open(test_filename).readlines()]
        box_lists = [[Box(object_reps, i) for i, object_reps in enumerate(box_rep)] for box_rep in data]
        self.executors = [NlvrExecutor(boxes) for boxes in box_lists]
        # y_loc increases as we go down from top to bottom, and x_loc from left to right. That is,
        # the origin is at the top-left corner.
        custom_rep = [[{"y_loc": 79, "size": 20, "type": "triangle", "x_loc": 27, "color": "Yellow"},
                       {"y_loc": 55, "size": 10, "type": "circle", "x_loc": 47, "color": "Black"}],
                      [{"y_loc": 44, "size": 30, "type": "square", "x_loc": 10, "color": "#0099ff"},
                       {"y_loc": 74, "size": 30, "type": "square", "x_loc": 40, "color": "Yellow"}],
                      [{"y_loc": 60, "size": 10, "type": "triangle", "x_loc": 12, "color": "#0099ff"}]]
        self.custom_executor = NlvrExecutor([Box(object_rep, i) for i, object_rep in
                                             enumerate(custom_rep)])

    def test_logical_form_with_assert_executes_correctly(self):
        executor = self.executors[0]
        # Utterance is "There is a circle closely touching a corner of a box." and label is "True".
        logical_form_true = "(object_count_greater_equals (touch_corner (circle (all_objects))) 1)"
        assert executor.execute(logical_form_true) is True
        logical_form_false = "(object_count_equals (touch_corner (circle (all_objects))) 0)"
        assert executor.execute(logical_form_false) is False

    def test_logical_form_with_box_filter_executes_correctly(self):
        executor = self.executors[2]
        # Utterance is "There is a box without a blue item." and label is "False".
        logical_form = "(box_exists (member_color_none_equals all_boxes color_blue))"
        assert executor.execute(logical_form) is False

    def test_logical_form_with_box_filter_within_object_filter_executes_correctly(self):
        executor = self.executors[2]
        # Utterance is "There are at least three blue items in boxes with blue items" and label
        # is "True".
        logical_form = "(object_count_greater_equals \
                            (object_in_box (member_color_any_equals all_boxes color_blue)) 3)"
        assert executor.execute(logical_form) is True

    def test_logical_form_with_same_color_executes_correctly(self):
        executor = self.executors[1]
        # Utterance is "There are exactly two blocks of the same color." and label is "True".
        logical_form = "(object_count_equals (same_color all_objects) 2)"
        assert executor.execute(logical_form) is True

    def test_logical_form_with_same_shape_executes_correctly(self):
        executor = self.executors[0]
        # Utterance is "There are less than three black objects of the same shape" and label is "False".
        logical_form = "(object_count_lesser (same_shape (black (all_objects))) 3)"
        assert executor.execute(logical_form) is False

    def test_logical_form_with_touch_wall_executes_correctly(self):
        executor = self.executors[0]
        # Utterance is "There are two black circles touching a wall" and label is "False".
        logical_form = "(object_count_greater_equals (touch_wall (black (circle (all_objects)))) 2)"
        assert executor.execute(logical_form) is False

    def test_logical_form_with_not_executes_correctly(self):
        executor = self.executors[2]
        # Utterance is "There are at most two medium triangles not touching a wall." and label is "True".
        logical_form = ("(object_count_lesser_equals ((negate_filter touch_wall) \
                                                                     (medium (triangle (all_objects)))) 2)")
        assert executor.execute(logical_form) is True

    def test_logical_form_with_color_comparison_executes_correctly(self):
        executor = self.executors[0]
        # Utterance is "The color of the circle touching the wall is black." and label is "True".
        logical_form = "(object_color_all_equals (circle (touch_wall (all_objects))) color_black)"
        assert executor.execute(logical_form) is True

    def test_spatial_relations_return_objects_in_the_same_box(self):
        # "above", "below", "top", "bottom" are relations defined only for objects within the same
        # box. So they should not return objects from other boxes.
        # Asserting that the color of the objects above the yellow triangle is only black (it is not
        # yellow or blue, which are colors of objects from other boxes)
        assert self.custom_executor.execute("(object_color_all_equals (above (yellow (triangle all_objects)))"
                                            " color_black)") is True
        # Asserting that the only shape below the blue square is a square.
        assert self.custom_executor.execute("(object_shape_all_equals (below (blue (square all_objects)))"
                                            " shape_square)") is True
        # Asserting the shape of the object at the bottom in the box with a circle is triangle.
        logical_form = ("(object_shape_all_equals (bottom (object_in_box"
                        " (member_shape_any_equals all_boxes shape_circle))) shape_triangle)")
        assert self.custom_executor.execute(logical_form) is True

        # Asserting the shape of the object at the top of the box with all squares is a square (!).
        logical_form = ("(object_shape_all_equals (top (object_in_box"
                        " (member_shape_all_equals all_boxes shape_square))) shape_square)")
        assert self.custom_executor.execute(logical_form) is True

    def test_touch_object_executes_correctly(self):
        # Assert that there is a yellow square touching a blue square.
        assert self.custom_executor.execute("(object_exists (yellow (square (touch_object (blue "
                                            "(square all_objects))))))") is True
        # Assert that the triangle does not touch the circle (they are out of vertical range).
        assert self.custom_executor.execute("(object_shape_none_equals (touch_object (triangle all_objects))"
                                            " shape_circle)") is True

    def test_spatial_relations_with_objects_from_different_boxes(self):
        # When the objects are from different boxes, top and bottom should return objects from
        # respective boxes.
        # There are triangles in two boxes, so top should return the top objects from both boxes.
        assert self.custom_executor.execute("(object_count_equals (top (object_in_box (member_shape_any_equals "
                                            "all_boxes shape_triangle))) 2)") is True

    def test_count_with_all_equals_throws_execution_error(self):
        # "*_all_equals" is a comparison valid only for sets (of colors and shapes). A comparison
        # with count should use "*_equals" instead.
        with self.assertRaises(ExecutionError):
            self.custom_executor.execute("(object_count_all_equals (top (object_in_box (member_shape_any_equals "
                                         "all_boxes shape_triangle))) 2)")

    def test_shape_with_equals_throws_execution_error(self):
        # "*_all_equals" is a comparison valid only for counts. A comparison with sets should use
        # "*_equals", "*_any_equals" or "*_non_equals" instead.
        with self.assertRaises(ExecutionError):
            self.custom_executor.execute("(object_shape_equals (top (object_in_box (member_shape_any_equals "
                                         "all_boxes shape_triangle))) shape_triangle)")

    def test_same_and_different_execute_correctly(self):
        # All the objects in the box with two objects of the same shape are squares.
        assert self.custom_executor.execute("(object_shape_all_equals "
                                            "(object_in_box (member_shape_same (member_count_equals all_boxes 2)))"
                                            " shape_square)") is True
        # There is a circle in the box with objects of different shapes.
        assert self.custom_executor.execute("(object_shape_any_equals (object_in_box "
                                            "(member_shape_different all_boxes)) shape_circle)") is True



"""
This module defines classes Object and Box (the two entities in the NLVR domain) and an NlvrWorld,
which mainly contains an execution method and related helper methods.
"""
from typing import List, Dict, Set
import logging

from nltk.sem.logic import Type
from overrides import overrides

from allennlp.common.util import JsonDict
from allennlp.semparse.type_declarations import nlvr_type_declaration as types
from allennlp.semparse.worlds.nlvr_box import Box
from allennlp.semparse.worlds.world import World
from allennlp.semparse.executors import NlvrExecutor

logger = logging.getLogger(__name__)  # pylint: disable=invalid-name


class NlvrWorld(World):
    """
    Class defining the world representation of NLVR. Defines an execution logic for logical forms
    in NLVR.  We just take the structured_rep from the JSON file to initialize this.

    Parameters
    ----------
    world_representation : ``JsonDict``
        structured_rep from the JSON file.
    """
    # pylint: disable=too-many-public-methods

    # When we're converting from logical forms to action sequences, this set tells us which
    # functions in the logical form are curried functions, and how many arguments the function
    # actually takes.  This is necessary because NLTK curries all multi-argument functions to a
    # series of one-argument function applications.  See `world._get_transitions` for more info.
    curried_functions = {
            types.BOX_COLOR_FILTER_TYPE: 2,
            types.BOX_SHAPE_FILTER_TYPE: 2,
            types.BOX_COUNT_FILTER_TYPE: 2,
            types.ASSERT_COLOR_TYPE: 2,
            types.ASSERT_SHAPE_TYPE: 2,
            types.ASSERT_BOX_COUNT_TYPE: 2,
            types.ASSERT_OBJECT_COUNT_TYPE: 2,
            }

    # TODO(pradeep): Define more spatial relationship methods: left_of, right_of..
    # They should be defined for objects within the same box.
    def __init__(self, world_representation: List[List[JsonDict]]) -> None:
        super(NlvrWorld, self).__init__(global_type_signatures=types.COMMON_TYPE_SIGNATURE,
                                        global_name_mapping=types.COMMON_NAME_MAPPING,
                                        num_nested_lambdas=0)
        boxes = set([Box(object_list, box_id) for box_id, object_list in
                     enumerate(world_representation)])
        self._executor = NlvrExecutor(boxes)

        # Mapping from terminal strings to productions that produce them.
        # Eg.: "yellow" -> "<o,o> -> yellow", "<b,<<b,e>,<e,b>>> -> filter_greater" etc.
        self.terminal_productions: Dict[str, str] = {}
        for constant in types.COMMON_NAME_MAPPING:
            alias = types.COMMON_NAME_MAPPING[constant]
            if alias in types.COMMON_TYPE_SIGNATURE:
                constant_type = types.COMMON_TYPE_SIGNATURE[alias]
                self.terminal_productions[constant] = "%s -> %s" % (constant_type, constant)

    @overrides
    def get_basic_types(self) -> Set[Type]:
        return types.BASIC_TYPES

    @overrides
    def get_valid_starting_types(self) -> Set[Type]:
        return {types.TRUTH_TYPE}

    def _get_curried_functions(self) -> Dict[Type, int]:
        return NlvrWorld.curried_functions

    @overrides
    def _map_name(self, name: str, keep_mapping: bool = False) -> str:
        return types.COMMON_NAME_MAPPING[name] if name in types.COMMON_NAME_MAPPING else name

    def get_agenda_for_sentence(self,
                                sentence: str,
                                add_paths_to_agenda: bool = False) -> List[str]:
        """
        Given a ``sentence``, returns a list of actions the sentence triggers as an ``agenda``. The
        ``agenda`` can be used while by a parser to guide the decoder.  sequences as possible. This
        is a simplistic mapping at this point, and can be expanded.

        Parameters
        ----------
        sentence : ``str``
            The sentence for which an agenda will be produced.
        add_paths_to_agenda : ``bool`` , optional
            If set, the agenda will also include nonterminal productions that lead to the terminals
            from the root node (default = False).
        """
        agenda = []
        sentence = sentence.lower()
        if sentence.startswith("there is a box") or sentence.startswith("there is a tower "):
            agenda.append(self.terminal_productions["box_exists"])
        elif sentence.startswith("there is a "):
            agenda.append(self.terminal_productions["object_exists"])

        if "<b,t> -> box_exists" not in agenda:
            # These are object filters and do not apply if we have a box_exists at the top.
            if "touch" in sentence:
                if "top" in sentence:
                    agenda.append(self.terminal_productions["touch_top"])
                elif "bottom" in sentence or "base" in sentence:
                    agenda.append(self.terminal_productions["touch_bottom"])
                elif "corner" in sentence:
                    agenda.append(self.terminal_productions["touch_corner"])
                elif "right" in sentence:
                    agenda.append(self.terminal_productions["touch_right"])
                elif "left" in sentence:
                    agenda.append(self.terminal_productions["touch_left"])
                elif "wall" in sentence or "edge" in sentence:
                    agenda.append(self.terminal_productions["touch_wall"])
                else:
                    agenda.append(self.terminal_productions["touch_object"])
            else:
                # The words "top" and "bottom" may be referring to top and bottom blocks in a tower.
                if "top" in sentence:
                    agenda.append(self.terminal_productions["top"])
                elif "bottom" in sentence or "base" in sentence:
                    agenda.append(self.terminal_productions["bottom"])

            if " not " in sentence:
                agenda.append(self.terminal_productions["negate_filter"])

        if " contains " in sentence or " has " in sentence:
            agenda.append(self.terminal_productions["all_boxes"])
        # This takes care of shapes, colors, top, bottom, big, small etc.
        for constant, production in self.terminal_productions.items():
            # TODO(pradeep): Deal with constant names with underscores.
            if "top" in constant or "bottom" in constant:
                # We already dealt with top, bottom, touch_top and touch_bottom above.
                continue
            if constant in sentence:
                if "<o,o> ->" in production and "<b,t> -> box_exists" in agenda:
                    if constant in ["square", "circle", "triangle"]:
                        agenda.append(self.terminal_productions[f"shape_{constant}"])
                    elif constant in ["yellow", "blue", "black"]:
                        agenda.append(self.terminal_productions[f"color_{constant}"])
                    else:
                        continue
                else:
                    agenda.append(production)
        # TODO (pradeep): Rules for "member_*" productions ("tower" or "box" followed by a color,
        # shape or number...)
        number_productions = self._get_number_productions(sentence)
        for production in number_productions:
            agenda.append(production)
        if not agenda:
            # None of the rules above was triggered!
            if "box" in sentence:
                agenda.append(self.terminal_productions["all_boxes"])
            else:
                agenda.append(self.terminal_productions["all_objects"])
        if add_paths_to_agenda:
            agenda = self._add_nonterminal_productions(agenda)
        return agenda

    @staticmethod
    def _get_number_productions(sentence: str) -> List[str]:
        """
        Gathers all the numbers in the sentence, and returns productions that lead to them.
        """
        # The mapping here is very simple and limited, which also shouldn't be a problem
        # because numbers seem to be represented fairly regularly.
        number_strings = {"one": "1", "two": "2", "three": "3", "four": "4", "five": "5", "six":
                          "6", "seven": "7", "eight": "8", "nine": "9", "ten": "10"}
        number_productions = []
        tokens = sentence.split()
        numbers = number_strings.values()
        for token in tokens:
            if token in numbers:
                number_productions.append(f"e -> {token}")
            elif token in number_strings:
                number_productions.append(f"e -> {number_strings[token]}")
        return number_productions

    def _add_nonterminal_productions(self, agenda: List[str]) -> List[str]:
        """
        Given a partially populated agenda with (mostly) terminal productions, this method adds the
        nonterminal productions that lead from the root to the terminal productions.
        """
        nonterminal_productions = set(agenda)
        for action in agenda:
            paths = self.get_paths_to_root(action, max_num_paths=5)
            for path in paths:
                for path_action in path:
                    nonterminal_productions.add(path_action)
        new_agenda = list(nonterminal_productions)
        return new_agenda

    def execute(self, logical_form: str) -> bool:
        return self._executor.execute(logical_form)



"""
We store the information related to context sensitive execution of logical forms here.
We assume that the logical forms are written in the variable-free language described in the paper
'Memory Augmented Policy Optimization for Program Synthesis with Generalization' by Liang et al.
The language is the main difference between this class and `WikiTablesWorld`. Also, this class defines
an executor for the variable-free logical forms.
"""
# TODO(pradeep): Merge this class with the `WikiTablesWorld` class, and move all the
# language-specific functionality into type declarations.
from typing import Dict, List, Set, Union
import re
import logging

from nltk.sem.logic import Type
from overrides import overrides

from allennlp.semparse.worlds.world import ParsingError, World
from allennlp.semparse.type_declarations import wikitables_variable_free as types
from allennlp.semparse.contexts import TableQuestionContext
from allennlp.semparse.executors import WikiTablesVariableFreeExecutor

logger = logging.getLogger(__name__)  # pylint: disable=invalid-name


class WikiTablesVariableFreeWorld(World):
    """
    World representation for the WikitableQuestions domain with the variable-free language used in
    the paper from Liang et al. (2018).

    Parameters
    ----------
    table_graph : ``TableQuestionKnowledgeGraph``
        Context associated with this world.
    """
    # When we're converting from logical forms to action sequences, this set tells us which
    # functions in the logical form are curried functions, and how many arguments the function
    # actually takes.  This is necessary because NLTK curries all multi-argument functions to a
    # series of one-argument function applications.  See `world._get_transitions` for more info.
    curried_functions = {
            types.SELECT_TYPE: 2,
            types.ROW_FILTER_WITH_GENERIC_COLUMN: 2,
            types.ROW_FILTER_WITH_COMPARABLE_COLUMN: 2,
            types.ROW_NUM_OP: 2,
            types.ROW_FILTER_WITH_COLUMN_AND_NUMBER: 3,
            types.ROW_FILTER_WITH_COLUMN_AND_DATE: 3,
            types.ROW_FILTER_WITH_COLUMN_AND_STRING: 3,
            types.NUM_DIFF_WITH_COLUMN: 3,
            }

    def __init__(self, table_context: TableQuestionContext) -> None:
        super().__init__(constant_type_prefixes={"string": types.STRING_TYPE,
                                                 "num": types.NUMBER_TYPE},
                         global_type_signatures=types.COMMON_TYPE_SIGNATURE,
                         global_name_mapping=types.COMMON_NAME_MAPPING)
        self.table_context = table_context
        # We add name mapping and signatures corresponding to specific column types to the local
        # name mapping based on the table content here.
        column_types = table_context.column_types.values()
        self._table_has_string_columns = False
        self._table_has_date_columns = False
        self._table_has_number_columns = False
        if "string" in column_types:
            for name, translated_name in types.STRING_COLUMN_NAME_MAPPING.items():
                signature = types.STRING_COLUMN_TYPE_SIGNATURE[translated_name]
                self._add_name_mapping(name, translated_name, signature)
            self._table_has_string_columns = True
        if "date" in column_types:
            for name, translated_name in types.DATE_COLUMN_NAME_MAPPING.items():
                signature = types.DATE_COLUMN_TYPE_SIGNATURE[translated_name]
                self._add_name_mapping(name, translated_name, signature)
            # Adding -1 to mapping because we need it for dates where not all three fields are
            # specified. We want to do this only when the table has a date column. This is because
            # the knowledge graph is also constructed in such a way that -1 is an entity with date
            # columns as the neighbors only if any date columns exist in the table.
            self._map_name(f"num:-1", keep_mapping=True)
            self._table_has_date_columns = True
        if "number" in column_types:
            for name, translated_name in types.NUMBER_COLUMN_NAME_MAPPING.items():
                signature = types.NUMBER_COLUMN_TYPE_SIGNATURE[translated_name]
                self._add_name_mapping(name, translated_name, signature)
            self._table_has_number_columns = True
        if "date" in column_types or "number" in column_types:
            for name, translated_name in types.COMPARABLE_COLUMN_NAME_MAPPING.items():
                signature = types.COMPARABLE_COLUMN_TYPE_SIGNATURE[translated_name]
                self._add_name_mapping(name, translated_name, signature)

        self.table_graph = table_context.get_table_knowledge_graph()

        self._executor = WikiTablesVariableFreeExecutor(self.table_context.table_data)

        # TODO (pradeep): Use a NameMapper for mapping entity names too.
        # For every new column name seen, we update this counter to map it to a new NLTK name.
        self._column_counter = 0

        # Adding entities and numbers seen in questions to the mapping.
        question_entities, question_numbers = table_context.get_entities_from_question()
        self._question_entities = [entity for entity, _ in question_entities]
        self._question_numbers = [number for number, _ in question_numbers]
        for entity in self._question_entities:
            # These entities all have prefix "string:"
            self._map_name(entity, keep_mapping=True)

        for number_in_question in self._question_numbers:
            self._map_name(f"num:{number_in_question}", keep_mapping=True)

        # Keeps track of column name productions so that we can add them to the agenda.
        self._column_productions_for_agenda: Dict[str, str] = {}

        # Adding column names to the local name mapping.
        for column_name, column_type in table_context.column_types.items():
            self._map_name(f"{column_type}_column:{column_name}", keep_mapping=True)

        self.terminal_productions: Dict[str, str] = {}
        name_mapping = [(name, mapping) for name, mapping in self.global_name_mapping.items()]
        name_mapping += [(name, mapping) for name, mapping in self.local_name_mapping.items()]
        signatures = self.global_type_signatures.copy()
        signatures.update(self.local_type_signatures)
        for predicate, mapped_name in name_mapping:
            if mapped_name in signatures:
                signature = signatures[mapped_name]
                self.terminal_productions[predicate] = f"{signature} -> {predicate}"

        # We don't need to recompute this ever; let's just compute it once and cache it.
        self._valid_actions: Dict[str, List[str]] = None

    @staticmethod
    def is_instance_specific_entity(entity_name: str) -> bool:
        """
        Instance specific entities are column names, strings and numbers. Returns True if the entity
        is one of those.
        """
        entity_is_number = False
        try:
            float(entity_name)
            entity_is_number = True
        except ValueError:
            pass
        # Column names start with "*_column:", strings start with "string:"
        return "_column:" in entity_name or entity_name.startswith("string:") or entity_is_number

    @overrides
    def _get_curried_functions(self) -> Dict[Type, int]:
        return WikiTablesVariableFreeWorld.curried_functions

    @overrides
    def get_basic_types(self) -> Set[Type]:
        basic_types = set(types.BASIC_TYPES)
        if self._table_has_string_columns:
            basic_types.add(types.STRING_COLUMN_TYPE)
        if self._table_has_date_columns:
            basic_types.add(types.DATE_COLUMN_TYPE)
            basic_types.add(types.COMPARABLE_COLUMN_TYPE)
        if self._table_has_number_columns:
            basic_types.add(types.NUMBER_COLUMN_TYPE)
            basic_types.add(types.COMPARABLE_COLUMN_TYPE)
        return basic_types

    @overrides
    def get_valid_starting_types(self) -> Set[Type]:
        return types.STARTING_TYPES

    def _translate_name_and_add_mapping(self, name: str) -> str:
        if "_column:" in name:
            # Column name
            translated_name = "C%d" % self._column_counter
            self._column_counter += 1
            if name.startswith("number_column:"):
                column_type = types.NUMBER_COLUMN_TYPE
            elif name.startswith("string_column:"):
                column_type = types.STRING_COLUMN_TYPE
            else:
                column_type = types.DATE_COLUMN_TYPE
            self._add_name_mapping(name, translated_name, column_type)
            self._column_productions_for_agenda[name] = f"{column_type} -> {name}"
        elif name.startswith("string:"):
            # We do not need to translate these names.
            translated_name = name
            self._add_name_mapping(name, translated_name, types.STRING_TYPE)
        elif name.startswith("num:"):
            # NLTK throws an error if it sees a "." in constants, which will most likely happen
            # within numbers as a decimal point. We're changing those to underscores.
            translated_name = name.replace(".", "_")
            if re.match("num:-[0-9_]+", translated_name):
                # The string is a negative number. This makes NLTK interpret this as a negated
                # expression and force its type to be TRUTH_VALUE (t).
                translated_name = translated_name.replace("-", "~")
            original_name = name.replace("num:", "")
            self._add_name_mapping(original_name, translated_name, types.NUMBER_TYPE)
        return translated_name

    @overrides
    def _map_name(self, name: str, keep_mapping: bool = False) -> str:
        if name not in types.COMMON_NAME_MAPPING and name not in self.local_name_mapping:
            if not keep_mapping:
                raise ParsingError(f"Encountered un-mapped name: {name}")
            translated_name = self._translate_name_and_add_mapping(name)
        else:
            if name in types.COMMON_NAME_MAPPING:
                translated_name = types.COMMON_NAME_MAPPING[name]
            else:
                translated_name = self.local_name_mapping[name]
        return translated_name

    def get_agenda(self):
        agenda_items = []
        question_tokens = [token.text for token in self.table_context.question_tokens]
        question = " ".join(question_tokens)
        if "at least" in question:
            agenda_items.append("filter_number_greater_equals")
        if "at most" in question:
            agenda_items.append("filter_number_lesser_equals")

        comparison_triggers = ["greater", "larger", "more"]
        if any("no %s than" %word in question for word in comparison_triggers):
            agenda_items.append("filter_number_lesser_equals")
        elif any("%s than" %word in question for word in comparison_triggers):
            agenda_items.append("filter_number_greater")
        for token in question_tokens:
            if token in ["next", "after", "below"]:
                agenda_items.append("next")
            if token in ["previous", "before", "above"]:
                agenda_items.append("previous")
            if token == "total":
                agenda_items.append("sum")
            if token == "difference":
                agenda_items.append("diff")
            if token == "average":
                agenda_items.append("average")
            if token in ["least", "smallest", "shortest", "lowest"] and "at least" not in question:
                # This condition is too brittle. But for most logical forms with "min", there are
                # semantically equivalent ones with "argmin". The exceptions are rare.
                if "what is the least" in question:
                    agenda_items.append("min")
                else:
                    agenda_items.append("argmin")
            if token in ["most", "largest", "highest", "longest", "greatest"] and "at most" not in question:
                # This condition is too brittle. But for most logical forms with "max", there are
                # semantically equivalent ones with "argmax". The exceptions are rare.
                if "what is the most" in question:
                    agenda_items.append("max")
                else:
                    agenda_items.append("argmax")
            if token in ["first", "top"]:
                agenda_items.append("first")
            if token in ["last", "bottom"]:
                agenda_items.append("last")

        if "how many" in question:
            if "sum" not in agenda_items and "average" not in agenda_items:
                # The question probably just requires counting the rows. But this is not very
                # accurate. The question could also be asking for a value that is in the table.
                agenda_items.append("count")
        agenda = []
        # Adding productions from the global set.
        for agenda_item in set(agenda_items):
            # Some agenda items may not be present in the terminal productions because some of these
            # terminals are table-content specific. For example, if the question triggered "sum",
            # and the table does not have number columns, we should not add "<r,<f,n>> -> sum" to
            # the agenda.
            if agenda_item in self.terminal_productions:
                agenda.append(self.terminal_productions[agenda_item])

        # Adding column names that occur in question.
        question_with_underscores = "_".join(question_tokens)
        normalized_question = re.sub("[^a-z0-9_]", "", question_with_underscores)
        # We keep track of tokens that are in column names being added to the agenda. We will not
        # add string productions to the agenda if those tokens were already captured as column
        # names.
        # Note: If the same string occurs multiple times, this may cause string productions being
        # omitted from the agenda unnecessarily. That is fine, as we want to err on the side of
        # adding fewer rules to the agenda.
        tokens_in_column_names: Set[str] = set()
        for column_name_with_type, signature in self._column_productions_for_agenda.items():
            column_name = column_name_with_type.split(":")[1]
            # Underscores ensure that the match is of whole words.
            if f"_{column_name}_" in normalized_question:
                agenda.append(signature)
                for token in column_name.split("_"):
                    tokens_in_column_names.add(token)

        # Adding all productions that lead to entities and numbers extracted from the question.
        for entity in self._question_entities:
            if entity.replace("string:", "") not in tokens_in_column_names:
                agenda.append(f"{types.STRING_TYPE} -> {entity}")

        for number in self._question_numbers:
            # The reason we check for the presence of the number in the question again is because
            # some of these numbers are extracted from number words like month names and ordinals
            # like "first". On looking at some agenda outputs, I found that they hurt more than help
            # in the agenda.
            if f"_{number}_" in normalized_question:
                agenda.append(f"{types.NUMBER_TYPE} -> {number}")
        return agenda


    def execute(self, logical_form: str) -> Union[List[str], int]:
        return self._executor.execute(logical_form)

    def evaluate_logical_form(self, logical_form: str, target_list: List[str]) -> bool:
        """
        Takes a logical forms and a list of target values as strings from the original lisp
        representation of instances, and returns True iff the logical form executes to those values.
        """
        return self._executor.evaluate_logical_form(logical_form, target_list)



import re

FILENAME = 'tutorials/tagger/basic_allennlp.py'

full_text = open(FILENAME).read()

comment_regex = r"[ ]*(####.*)\n"

parts = re.split(comment_regex, full_text)

# remove newlines, then triple quotes, then more newlines
HTML = parts[0].strip().strip('"""').strip()

parts = parts[1:]
num_parts = len(parts) // 2

comments = parts[::2]
codes = parts[1::2]

HTML += """<div id="annotated-code">
  <!-- Code Blocks -->
  <div class="annotated-code__pane annotated-code__pane--code-container">
"""

for i, code in enumerate(codes):
    code = code.rstrip("\n")
    HTML += f"""<div class="annotated-code__code-block" id="c{i}">
{{% highlight python %}}
{code}
{{% endhighlight %}}
</div>
"""

HTML += """</div>
    <!-- END Code Blocks -->

    <!-- Annotations -->
    <div class="annotated-code__pane annotated-code__pane--annotations-container">
        <ul id="annotated-code__annotations">
"""

for i, comment in enumerate(comments):
    comment = comment.strip("####").strip()
    HTML += f"""<li class="annotation" id="a{i}">{comment}</li>
"""

HTML += """</ul>
  </div><!-- END Annotations -->
</div><!-- END Annotated Code -->
 {% include more-tutorials.html %}
"""

with open('tutorials/tagger/default-tutorial.html', 'w') as f:
    f.write(HTML)



from typing import Dict, List, Tuple, Union

import torch
import numpy as np

from allennlp.common.checks import ConfigurationError
from allennlp.data.vocabulary import Vocabulary
from allennlp.models.model import Model
from allennlp.modules.text_field_embedders import TextFieldEmbedder
from allennlp.modules.sampled_softmax_loss import SampledSoftmaxLoss
from allennlp.modules.seq2seq_encoders import Seq2SeqEncoder
from allennlp.nn.util import get_text_field_mask
from allennlp.nn import InitializerApplicator


class _SoftmaxLoss(torch.nn.Module):
    """
    Given some embeddings and some targets, applies a linear layer
    to create logits over possible words and then returns the
    negative log likelihood.
    """
    def __init__(self,
                 num_words: int,
                 embedding_dim: int) -> None:
        super().__init__()

        # TODO(joelgrus): implement tie_embeddings (maybe)
        self.tie_embeddings = False

        self.softmax_w = torch.nn.Parameter(
                torch.randn(embedding_dim, num_words) / np.sqrt(embedding_dim)
        )
        self.softmax_b = torch.nn.Parameter(torch.zeros(num_words))

    def forward(self, embeddings: torch.Tensor, targets: torch.Tensor) -> torch.Tensor:
        # pylint: disable=arguments-differ
        # embeddings is size (n, embedding_dim)
        # targets is (batch_size, ) with the correct class id
        # Does not do any count normalization / divide by batch size
        probs = torch.nn.functional.log_softmax(
                torch.matmul(embeddings, self.softmax_w) + self.softmax_b,
                dim=-1
        )

        return torch.nn.functional.nll_loss(probs, targets.long(), reduction="sum")


@Model.register('bidirectional-language-model')
@Model.register('bidirectional_language_model')
class BidirectionalLanguageModel(Model):
    """
    The ``BidirectionalLanguageModel`` applies a bidirectional "contextualizing"
    ``Seq2SeqEncoder`` to uncontextualized embeddings, using a ``SoftmaxLoss``
    module (defined above) to compute the language modeling loss.

    It is IMPORTANT that your bidirectional ``Seq2SeqEncoder`` does not do any
    "peeking ahead". That is, for its forward direction it should only consider
    embeddings at previous timesteps, and for its backward direction only embeddings
    at subsequent timesteps. If this condition is not met, your language model is
    cheating.

    Parameters
    ----------
    vocab: ``Vocabulary``
    text_field_embedder: ``TextFieldEmbedder``
        Used to embed the indexed tokens we get in ``forward``.
    contextualizer: ``Seq2SeqEncoder``
        Used to "contextualize" the embeddings. As described above,
        this encoder must not cheat by peeking ahead.
    dropout: ``float``, optional (default: None)
        If specified, dropout is applied to the contextualized embeddings before computation of
        the softmax. The contextualized embeddings themselves are returned without dropout.
    loss_scale: ``Union[float, str]``, optional (default: 1.0)
        This scaling factor is applied to the average language model loss.
        You can also specify ``"n_samples"`` in which case we compute total
        loss across all predictions.
    num_samples: ``int``, optional (default: None)
        If provided, the model will use ``SampledSoftmaxLoss``
        with the specified number of samples. Otherwise, it will use
        the full ``_SoftmaxLoss`` defined above.
    sparse_embeddings: ``bool``, optional (default: False)
        Passed on to ``SampledSoftmaxLoss`` if True.
    """
    def __init__(self,
                 vocab: Vocabulary,
                 text_field_embedder: TextFieldEmbedder,
                 contextualizer: Seq2SeqEncoder,
                 dropout: float = None,
                 loss_scale: Union[float, str] = 1.0,
                 num_samples: int = None,
                 sparse_embeddings: bool = False,
                 initializer: InitializerApplicator = None) -> None:
        super().__init__(vocab)
        self._text_field_embedder = text_field_embedder

        if not contextualizer.is_bidirectional():
            raise ConfigurationError("contextualizer must be bidirectional")

        self._contextualizer = contextualizer
        # The dimension for making predictions just in the forward
        # (or backward) direction.
        self._forward_dim = contextualizer.get_output_dim() // 2

        # TODO(joelgrus): more sampled softmax configuration options, as needed.
        if num_samples is not None:
            self._softmax_loss = SampledSoftmaxLoss(num_words=vocab.get_vocab_size(),
                                                    embedding_dim=self._forward_dim,
                                                    num_samples=num_samples,
                                                    sparse=sparse_embeddings)
        else:
            self._softmax_loss = _SoftmaxLoss(num_words=vocab.get_vocab_size(),
                                              embedding_dim=self._forward_dim)

        # TODO(brendanr): Output perplexity here. e^loss
        self.register_buffer('_last_average_loss', torch.zeros(1))

        if dropout:
            self._dropout = torch.nn.Dropout(dropout)
        else:
            self._dropout = lambda x: x

        self._loss_scale = loss_scale
        if initializer is not None:
            initializer(self)

    def _get_target_token_embedding(self,
                                    token_embeddings: torch.Tensor,
                                    mask: torch.Tensor,
                                    direction: int) -> torch.Tensor:
        # Need to shift the mask in the correct direction
        zero_col = token_embeddings.new_zeros(mask.size(0), 1).byte()
        if direction == 0:
            # forward direction, get token to right
            shifted_mask = torch.cat([zero_col, mask[:, 0:-1]], dim=1)
        else:
            shifted_mask = torch.cat([mask[:, 1:], zero_col], dim=1)
        return token_embeddings.masked_select(shifted_mask.unsqueeze(-1)).view(-1, self._forward_dim)

    def _compute_loss(self,
                      lm_embeddings: torch.Tensor,
                      token_embeddings: torch.Tensor,
                      forward_targets: torch.Tensor,
                      backward_targets: torch.Tensor) -> Tuple[torch.Tensor, torch.Tensor]:
        # lm_embeddings is shape (batch_size, timesteps, dim * 2)
        # forward_targets, backward_targets are shape (batch_size, timesteps)
        # masked with 0
        forward_embeddings, backward_embeddings = lm_embeddings.chunk(2, -1)
        losses: List[torch.Tensor] = []
        for idx, embedding, targets in ((0, forward_embeddings, forward_targets),
                                        (1, backward_embeddings, backward_targets)):
            mask = targets > 0
            # we need to subtract 1 to undo the padding id since the softmax
            # does not include a padding dimension

            # shape (batch_size * timesteps, )
            non_masked_targets = targets.masked_select(mask) - 1

            # shape (batch_size * timesteps, embedding_dim)
            non_masked_embedding = embedding.masked_select(
                    mask.unsqueeze(-1)
            ).view(-1, self._forward_dim)
            # note: need to return average loss across forward and backward
            # directions, but total sum loss across all batches.
            # Assuming batches include full sentences, forward and backward
            # directions have the same number of samples, so sum up loss
            # here then divide by 2 just below
            if not self._softmax_loss.tie_embeddings or not self._use_character_inputs:
                losses.append(self._softmax_loss(non_masked_embedding, non_masked_targets))
            else:
                # we also need the token embeddings corresponding to the
                # the targets
                raise NotImplementedError("This requires SampledSoftmaxLoss, which isn't implemented yet.")
                # pylint: disable=unreachable
                non_masked_token_embedding = self._get_target_token_embedding(token_embeddings, mask, idx)
                losses.append(self._softmax(non_masked_embedding,
                                            non_masked_targets,
                                            non_masked_token_embedding))

        return losses[0], losses[1]

    def delete_softmax(self) -> None:
        """
        Remove the softmax weights. Useful for saving memory when calculating the loss
        is not necessary, e.g. in an embedder.
        """
        self._softmax_loss = None

    def num_layers(self) -> int:
        """
        Returns the depth of this LM. That is, how many layers the contextualizer has plus one for
        the non-contextual layer.
        """
        if hasattr(self._contextualizer, 'num_layers'):
            return self._contextualizer.num_layers + 1
        else:
            raise NotImplementedError(f"Contextualizer of type {type(self._contextualizer)} " +
                                      "does not report how many layers it has.")

    def forward(self,  # type: ignore
                source: Dict[str, torch.LongTensor]) -> Dict[str, torch.Tensor]:
        """
        Computes the averaged forward and backward LM loss from the batch.

        By convention, the input dict is required to have at least a ``"tokens"``
        entry that's the output of a ``SingleIdTokenIndexer``, which is used
        to compute the language model targets.

        Parameters
        ----------
        tokens: ``torch.Tensor``, required.
            The output of ``Batch.as_tensor_dict()`` for a batch of sentences.

        Returns
        -------
        Dict with keys:

        ``'loss'``: ``torch.Tensor``
            averaged forward/backward negative log likelihood
        ``'forward_loss'``: ``torch.Tensor``
            forward direction negative log likelihood
        ``'backward_loss'``: ``torch.Tensor``
            backward direction negative log likelihood
        ``'lm_embeddings'``: ``Union[torch.Tensor, List[torch.Tensor]]``
            (batch_size, timesteps, embed_dim) tensor of top layer contextual representations or
            list of all layers. No dropout applied.
        ``'noncontextual_token_embeddings'``: ``torch.Tensor``
            (batch_size, timesteps, token_embed_dim) tensor of bottom layer noncontextual
            representations
        ``'mask'``: ``torch.Tensor``
            (batch_size, timesteps) mask for the embeddings
        """
        # pylint: disable=arguments-differ
        mask = get_text_field_mask(source)

        # shape (batch_size, timesteps, embedding_size)
        embeddings = self._text_field_embedder(source)

        # Either the top layer or all layers.
        contextual_embeddings: Union[torch.Tensor, List[torch.Tensor]] = self._contextualizer(
                embeddings, mask
        )

        return_dict = {}

        # If we have target tokens, calculate the loss.
        token_ids = source.get("tokens")
        if token_ids is not None:
            assert isinstance(contextual_embeddings, torch.Tensor)

            # Use token_ids to compute targets
            forward_targets = torch.zeros_like(token_ids)
            backward_targets = torch.zeros_like(token_ids)
            forward_targets[:, 0:-1] = token_ids[:, 1:]
            backward_targets[:, 1:] = token_ids[:, 0:-1]

            # add dropout
            contextual_embeddings_with_dropout = self._dropout(contextual_embeddings)

            # compute softmax loss
            forward_loss, backward_loss = self._compute_loss(contextual_embeddings_with_dropout,
                                                             embeddings,
                                                             forward_targets,
                                                             backward_targets)

            num_targets = torch.sum((forward_targets > 0).long())
            if num_targets > 0:
                average_loss = 0.5 * (forward_loss + backward_loss) / num_targets.float()
            else:
                average_loss = torch.tensor(0.0).to(forward_targets.device)  # pylint: disable=not-callable
            # this is stored to compute perplexity if needed
            self._last_average_loss[0] = average_loss.detach().item()

            if num_targets > 0:
                # loss is directly minimized
                if self._loss_scale == 'n_samples':
                    scale_factor = num_targets.float()
                else:
                    scale_factor = self._loss_scale

                return_dict.update({
                        'loss': average_loss * scale_factor,
                        'forward_loss': forward_loss * scale_factor / num_targets.float(),
                        'backward_loss': backward_loss * scale_factor / num_targets.float()
                })
            else:
                # average_loss zero tensor, return it for all
                return_dict.update({
                        'loss': average_loss,
                        'forward_loss': average_loss,
                        'backward_loss': average_loss
                })

        return_dict.update({
                # Note: These embeddings do not have dropout applied.
                'lm_embeddings': contextual_embeddings,
                'noncontextual_token_embeddings': embeddings,
                'mask': mask
        })

        return return_dict



# pylint: disable=no-self-use,invalid-name
import json

from allennlp.common.testing import AllenNlpTestCase
from allennlp.semparse.worlds.nlvr_world import NlvrWorld


class TestNlvrWorld(AllenNlpTestCase):
    def setUp(self):
        super().setUp()
        test_filename = self.FIXTURES_ROOT / "data" / "nlvr" / "sample_ungrouped_data.jsonl"
        data = [json.loads(line)["structured_rep"] for line in open(test_filename).readlines()]
        self.worlds = [NlvrWorld(rep) for rep in data]

    def test_get_action_sequence_removes_currying_for_all_nlvr_functions(self):
        world = self.worlds[0]
        # box_color_filter
        logical_form = "(member_color_all_equals all_boxes color_blue)"
        parsed_logical_form = world.parse_logical_form(logical_form)
        action_sequence = world.get_action_sequence(parsed_logical_form)
        assert 'b -> [<b,<c,b>>, b, c]' in action_sequence

        # box_shape_filter
        logical_form = "(member_shape_all_equals all_boxes shape_square)"
        parsed_logical_form = world.parse_logical_form(logical_form)
        action_sequence = world.get_action_sequence(parsed_logical_form)
        assert 'b -> [<b,<s,b>>, b, s]' in action_sequence

        # box_count_filter
        logical_form = "(member_count_equals all_boxes 3)"
        parsed_logical_form = world.parse_logical_form(logical_form)
        action_sequence = world.get_action_sequence(parsed_logical_form)
        assert 'b -> [<b,<e,b>>, b, e]' in action_sequence

        # assert_color
        logical_form = "(object_color_all_equals all_objects color_blue)"
        parsed_logical_form = world.parse_logical_form(logical_form)
        action_sequence = world.get_action_sequence(parsed_logical_form)
        assert 't -> [<o,<c,t>>, o, c]' in action_sequence

        # assert_shape
        logical_form = "(object_shape_all_equals all_objects shape_square)"
        parsed_logical_form = world.parse_logical_form(logical_form)
        action_sequence = world.get_action_sequence(parsed_logical_form)
        assert 't -> [<o,<s,t>>, o, s]' in action_sequence

        # assert_box_count
        logical_form = "(box_count_equals all_boxes 1)"
        parsed_logical_form = world.parse_logical_form(logical_form)
        action_sequence = world.get_action_sequence(parsed_logical_form)
        assert 't -> [<b,<e,t>>, b, e]' in action_sequence

        # assert_object_count
        logical_form = "(object_count_equals all_objects 1)"
        parsed_logical_form = world.parse_logical_form(logical_form)
        action_sequence = world.get_action_sequence(parsed_logical_form)
        assert 't -> [<o,<e,t>>, o, e]' in action_sequence

    def test_logical_form_with_object_filter_returns_correct_action_sequence(self):
        nlvr_world = self.worlds[0]
        logical_form = "(object_color_all_equals (circle (touch_wall (all_objects))) color_black)"
        expression = nlvr_world.parse_logical_form(logical_form)
        action_sequence = nlvr_world.get_action_sequence(expression)
        assert action_sequence == ['@start@ -> t', 't -> [<o,<c,t>>, o, c]',
                                   '<o,<c,t>> -> object_color_all_equals', 'o -> [<o,o>, o]',
                                   '<o,o> -> circle', 'o -> [<o,o>, o]', '<o,o> -> touch_wall',
                                   'o -> all_objects', 'c -> color_black']

    def test_logical_form_with_box_filter_returns_correct_action_sequence(self):
        nlvr_world = self.worlds[0]
        logical_form = "(box_exists (member_color_none_equals all_boxes color_blue))"
        expression = nlvr_world.parse_logical_form(logical_form)
        action_sequence = nlvr_world.get_action_sequence(expression)
        assert action_sequence == ['@start@ -> t', 't -> [<b,t>, b]', '<b,t> -> box_exists',
                                   'b -> [<b,<c,b>>, b, c]', '<b,<c,b>> -> member_color_none_equals',
                                   'b -> all_boxes', 'c -> color_blue']

    def test_get_agenda_for_sentence(self):
        world = self.worlds[0]
        agenda = world.get_agenda_for_sentence("there is a tower with exactly two yellow blocks")
        assert set(agenda) == set(['c -> color_yellow', '<b,t> -> box_exists', 'e -> 2'])
        agenda = world.get_agenda_for_sentence("There is at most one yellow item closely touching "
                                               "the bottom of a box.")
        assert set(agenda) == set(['<o,o> -> yellow', '<o,o> -> touch_bottom', 'e -> 1'])
        agenda = world.get_agenda_for_sentence("There is at most one yellow item closely touching "
                                               "the right wall of a box.")
        assert set(agenda) == set(['<o,o> -> yellow', '<o,o> -> touch_right', 'e -> 1'])
        agenda = world.get_agenda_for_sentence("There is at most one yellow item closely touching "
                                               "the left wall of a box.")
        assert set(agenda) == set(['<o,o> -> yellow', '<o,o> -> touch_left', 'e -> 1'])
        agenda = world.get_agenda_for_sentence("There is at most one yellow item closely touching "
                                               "a wall of a box.")
        assert set(agenda) == set(['<o,o> -> yellow', '<o,o> -> touch_wall', 'e -> 1'])
        agenda = world.get_agenda_for_sentence("There is exactly one square touching any edge")
        assert set(agenda) == set(['<o,o> -> square', '<o,o> -> touch_wall', 'e -> 1'])
        agenda = world.get_agenda_for_sentence("There is exactly one square not touching any edge")
        assert set(agenda) == set(['<o,o> -> square', '<o,o> -> touch_wall', 'e -> 1',
                                   '<<o,o>,<o,o>> -> negate_filter'])
        agenda = world.get_agenda_for_sentence("There is only 1 tower with 1 blue block at the base")
        assert set(agenda) == set(['<o,o> -> blue', 'e -> 1', '<o,o> -> bottom', 'e -> 1'])
        agenda = world.get_agenda_for_sentence("There is only 1 tower that has 1 blue block at the top")
        assert set(agenda) == set(['<o,o> -> blue', 'e -> 1', '<o,o> -> top', 'e -> 1',
                                   'b -> all_boxes'])
        agenda = world.get_agenda_for_sentence("There is exactly one square touching the blue "
                                               "triangle")
        assert set(agenda) == set(['<o,o> -> square', '<o,o> -> blue', '<o,o> -> triangle',
                                   '<o,o> -> touch_object', 'e -> 1'])

    def test_get_agenda_for_sentence_correctly_adds_object_filters(self):
        # In logical forms that contain "box_exists" at the top, there can never be object filtering
        # operations like "blue", "square" etc. In those cases, strings like "blue" and "square" in
        # sentences should map to "color_blue" and "shape_square" respectively.
        world = self.worlds[0]
        agenda = world.get_agenda_for_sentence("there is a box with exactly two yellow triangles "
                                               "touching the top edge")
        assert "<o,o> -> yellow" not in agenda
        assert "c -> color_yellow" in agenda
        assert "<o,o> -> triangle" not in agenda
        assert "s -> shape_triangle" in agenda
        assert "<o,o> -> touch_top" not in agenda
        agenda = world.get_agenda_for_sentence("there are exactly two yellow triangles touching the"
                                               " top edge")
        assert "<o,o> -> yellow" in agenda
        assert "c -> color_yellow" not in agenda
        assert "<o,o> -> triangle" in agenda
        assert "s -> shape_triangle" not in agenda
        assert "<o,o> -> touch_top" in agenda



from typing import List

from allennlp.common.util import JsonDict
from allennlp.semparse.worlds.nlvr_object import Object


class Box:
    """
    This class represents each box containing objects in NLVR.

    Parameters
    ----------
    objects_list : ``List[JsonDict]``
        List of objects in the box, as given by the json file.
    box_id : ``int``
        An integer identifying the box index (0, 1 or 2).
    """
    def __init__(self,
                 objects_list: List[JsonDict],
                 box_id: int) -> None:
        self._name = f"box {box_id + 1}"
        self._objects_string = str([str(_object) for _object in objects_list])
        self.objects = set([Object(object_dict, self._name) for object_dict in objects_list])

    def __str__(self):
        return self._objects_string

    def __hash__(self):
        return hash(str(self))

    def __eq__(self, other):
        return str(self) == str(other)



# pylint: disable=invalid-name,arguments-differ,abstract-method
import numpy as np

from allennlp.common.testing import ModelTestCase

class TestBidirectionalLM(ModelTestCase):
    def setUp(self):
        super().setUp()
        self.set_up_model(self.FIXTURES_ROOT / 'bidirectional_lm' / 'experiment.jsonnet',
                          self.FIXTURES_ROOT / 'bidirectional_lm' / 'sentences.txt')

    # pylint: disable=no-member
    def test_bidirectional_lm_can_train_save_and_load(self):
        self.ensure_model_can_train_save_and_load(self.param_file)

    def test_batch_predictions_are_consistent(self):
        self.ensure_batch_predictions_are_consistent()

    def test_forward_pass_runs_correctly(self):
        training_tensors = self.dataset.as_tensor_dict()
        result = self.model(**training_tensors)

        assert set(result) == {"loss", "forward_loss", "backward_loss",
                               "lm_embeddings", "noncontextual_token_embeddings", "mask"}

        # The model should preserve the BOS / EOS tokens.
        embeddings = result["lm_embeddings"]
        assert tuple(embeddings.shape) == (2, 8, 14)

        loss = result["loss"].item()
        forward_loss = result["forward_loss"].item()
        backward_loss = result["backward_loss"].item()

        np.testing.assert_almost_equal(loss, (forward_loss + backward_loss) / 2, decimal=3)


class TestBidirectionalLMUnsampled(TestBidirectionalLM):
    def setUp(self):
        super().setUp()
        self.set_up_model(self.FIXTURES_ROOT / 'bidirectional_lm' / 'experiment_unsampled.jsonnet',
                          self.FIXTURES_ROOT / 'bidirectional_lm' / 'sentences.txt')

class TestBidirectionalLMTransformer(TestBidirectionalLM):
    def setUp(self):
        super().setUp()
        self.set_up_model(self.FIXTURES_ROOT / 'bidirectional_lm' / 'experiment_transformer.jsonnet',
                          self.FIXTURES_ROOT / 'bidirectional_lm' / 'sentences.txt')

    def test_forward_pass_runs_correctly(self):
        training_tensors = self.dataset.as_tensor_dict()
        result = self.model(**training_tensors)

        assert set(result) == {"loss", "forward_loss", "backward_loss",
                               "lm_embeddings", "noncontextual_token_embeddings", "mask"}

        # The model should preserve the BOS / EOS tokens.
        embeddings = result["lm_embeddings"]
        # The BidirectionalLanguageModelTransformer uses input size * 2 as the output size unlike
        # a bidirectional LSTM, which uses hidden size * 2.
        assert tuple(embeddings.shape) == (2, 8, 32)

        loss = result["loss"].item()
        forward_loss = result["forward_loss"].item()
        backward_loss = result["backward_loss"].item()

        np.testing.assert_almost_equal(loss, (forward_loss + backward_loss) / 2, decimal=3)
